const char *diagnostics_status_success = "Success";
const char *diagnostics_status_error_dnsservernotavailable = "Error_DNSServerNotAvailable";
const char *diagnostics_status_error_hostnamenotresolved = "Error_HostNameNotResolved";
const char *diagnostics_status_error_timeout = "Error_Timeout";
const char *diagnostics_status_error_other = "Error_Other";

const char *diagnostics_answertype_none = "None";
const char *diagnostics_answertype_authoritative = "Authoritative";
const char *diagnostics_answertype_nonauthoritative = "NonAuthoritative";

unsigned int timeout_us = 0;
bool has_diagnostics = false;
struct dig_lookup_result lookup_result;

bool bind_interface(const char *interface)
{
	struct ifreq ifr;
	int fd, ret;
	if (!interface)
		return false;
	if (have_ipv4) {
		fd = socket(AF_INET, SOCK_DGRAM, 0);
		if (fd < 0)
			return false;
		ifr.ifr_addr.sa_family = AF_INET;
		strncpy(ifr.ifr_name, interface, IFNAMSIZ);
		ret = ioctl(fd, SIOCGIFADDR, &ifr);
		close(fd);
		if (ret == 0) {
			isc_sockaddr_fromin(&bind_address, &((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr, 0);
			isc_net_disableipv6();
			return true;
		}
	}
	if (have_ipv6) {
		fd = socket(AF_INET6, SOCK_DGRAM, 0);
		if (fd < 0)
			return false;
		ifr.ifr_addr.sa_family = AF_INET6;
		strncpy(ifr.ifr_name, interface, IFNAMSIZ);
		ret = ioctl(fd, SIOCGIFADDR, &ifr);
		close(fd);
		if (ret == 0) {
			isc_sockaddr_fromin6(&bind_address, &((struct sockaddr_in6 *)&ifr.ifr_addr)->sin6_addr, 0);
			isc_net_disableipv4();
			return true;
		}
	}
	return false;
}

void set_diagnostics_status(const char *status)
{
	if (has_diagnostics && status && strcmp(lookup_result.Status, diagnostics_status_success) && strcmp(lookup_result.Status, diagnostics_status_error_dnsservernotavailable))
		lookup_result.Status = status;
}

void set_diagnostics_answertype(const char *answertype)
{
	if (has_diagnostics && answertype)
		lookup_result.AnswerType = answertype;
}

void set_diagnostics_hostnamereturned(const char *hostnamereturned)
{
	if (has_diagnostics && hostnamereturned && !lookup_result.HostNameReturned)
		lookup_result.HostNameReturned = strdup(hostnamereturned);
}

void add_diagnostics_ipaddress(const char *ipaddress, int length)
{
	if (has_diagnostics && ipaddress && length > 0) {
		int offset = 0;
		if (lookup_result.IPAddresses) {
			offset = strlen(lookup_result.IPAddresses);
			lookup_result.IPAddresses = realloc(lookup_result.IPAddresses, offset + length + 2);
			lookup_result.IPAddresses[offset] = ',';
			++offset;
		} else {
			lookup_result.IPAddresses = malloc(length + 1);
		}
		memcpy(lookup_result.IPAddresses + offset, ipaddress, length);
		lookup_result.IPAddresses[offset + length] = 0;
	}
}

void set_diagnostics_dnsserverip(const isc_sockaddr_t *sockaddr)
{
	if (has_diagnostics && sockaddr) {
		isc_result_t result;
		isc_buffer_t buf;
		char array[256] = {0};
		isc_buffer_init(&buf, array, 256);
		result = isc_sockaddr_totext(sockaddr, &buf);
		if (result == ISC_R_SUCCESS) {
			char *hex = strstr(array, "#");
			if (hex)
				*hex = 0;
			lookup_result.DNSServerIP = strdup(array);
		}
	}
}

void set_diagnostics_responsetime(int responsetime)
{
	static int first = 1;
	if (has_diagnostics && strcmp(lookup_result.Status, diagnostics_status_error_timeout) && first) {
		first = 0;
		lookup_result.ResponseTime = responsetime;
	}
}

void init_diagnostics_result(void)
{
	memset(&lookup_result, 0, sizeof(lookup_result));
	lookup_result.Status = diagnostics_status_error_other;
	lookup_result.AnswerType = diagnostics_answertype_none;
}

void print_diagnostics_result(void)
{
	if (has_diagnostics) {
		printf("NSLookupDiagnosticsResult#Status=%s#AnswerType=%s#HostNameReturned=%s#IPAddresses=%s#DNSServerIP=%s#ResponseTime=%u\n",
		       lookup_result.Status, lookup_result.AnswerType, lookup_result.HostNameReturned ?: "", lookup_result.IPAddresses ?: "", lookup_result.DNSServerIP ?: "", lookup_result.ResponseTime/1000);
		if (lookup_result.HostNameReturned)
			free(lookup_result.HostNameReturned);
		if (lookup_result.IPAddresses)
			free(lookup_result.IPAddresses);
		if (lookup_result.DNSServerIP)
			free(lookup_result.DNSServerIP);
	}
}
